import React, { createContext, useState, useRef, useContext, useMemo, useEffect } from 'react';

function removeKeysFromObject(obj, keys) {
    return Object.keys(obj)
        .filter((key) => !keys.includes(key))
        .reduce((acc, key) => {
        acc[key] = obj[key];
        return acc;
    }, {});
}

const baseSerializeQueryParam = (value) => {
    if (isInvalidValue(value)) {
        return undefined;
    }
    if (Array.isArray(value)) {
        return value
            .filter((v) => !isInvalidValue(v))
            .map((v) => (typeof v === 'object' ? JSON.stringify(v) : v.toString()));
    }
    if (value instanceof Date) {
        return value.toISOString();
    }
    if (typeof value === 'object') {
        return JSON.stringify(value, function (_key, value) {
            if (typeof value === 'number' && isNaN(value)) {
                return 'NaN';
            }
            if (value === Infinity) {
                return 'Infinity';
            }
            if (value === -Infinity) {
                return '-Infinity';
            }
            if (value !== null && isInvalidValue(value)) {
                return undefined;
            }
            if (typeof value === 'object' &&
                !isEmptyObject(value) &&
                JSON.stringify(value) === '{}') {
                return undefined;
            }
            return value;
        });
    }
    return value === null || value === void 0 ? void 0 : value.toString();
};
function isInvalidValue(value) {
    return (value === null ||
        value === undefined ||
        (typeof value === 'number' && isNaN(value)) ||
        typeof value === 'symbol' ||
        typeof value === 'function' ||
        typeof value === 'bigint');
}
function isEmptyObject(value) {
    if (typeof value !== 'object' || value === null || Array.isArray(value)) {
        return false;
    }
    if (Object.prototype.toString.call(value) !== '[object Object]') {
        return false;
    }
    return Object.keys(value).length === 0;
}

function serializeQueryParamStates(state, customSerializeQueryParam) {
    const queryParams = {};
    for (const [key, value] of Object.entries(state)) {
        let serializedValue;
        if (value.serialize) {
            serializedValue = value.serialize(value.value);
        }
        else {
            serializedValue = customSerializeQueryParam
                ? customSerializeQueryParam(value.value)
                : baseSerializeQueryParam(value.value);
        }
        if (serializedValue !== undefined) {
            queryParams[key] = serializedValue;
        }
    }
    return queryParams;
}

function areUrlQueryValuesEqual(a, b) {
    if (Array.isArray(a) && Array.isArray(b)) {
        return a.length === b.length && a.every((value, index) => value === b[index]);
    }
    else if (typeof a === 'string' && typeof b === 'string') {
        return a === b;
    }
    return false;
}

function areUrlQueriesEqual(a, b) {
    const aKeys = Object.keys(a);
    const bKeys = Object.keys(b);
    if (aKeys.length !== bKeys.length || !aKeys.every((key) => bKeys.includes(key))) {
        return false;
    }
    return aKeys.every((key) => {
        const aValue = a[key];
        const bValue = b[key];
        return areUrlQueryValuesEqual(aValue, bValue);
    });
}

const NextQueryParamsContext = createContext(null);
const NextQueryParamsProvider = ({ children, adapter }) => {
    return (React.createElement(NextQueryParamsContext.Provider, { value: {
            adapter
        } }, children));
};

function getChangedUrlQueryKeys(a, b) {
    const aKeys = Object.keys(a);
    const bKeys = Object.keys(b);
    const allKeys = Array.from(new Set([...aKeys, ...bKeys]));
    return allKeys.filter((key) => {
        const aValue = a[key];
        const bValue = b[key];
        return (aValue === undefined || bValue === undefined || !areUrlQueryValuesEqual(aValue, bValue));
    });
}

function useNextQueryParams(params, adapter) {
    const [init, setInit] = useState(true);
    const [prevUrlQuery, setPrevUrlQuery] = useState({});
    const prevHasUrlQueryChanged = useRef(false);
    const keys = Object.keys(params);
    const nextQueryParamsContextValues = useContext(NextQueryParamsContext);
    const contextAdapter = nextQueryParamsContextValues === null || nextQueryParamsContextValues === void 0 ? void 0 : nextQueryParamsContextValues.adapter;
    const isRouterReady = (adapter === null || adapter === void 0 ? void 0 : adapter.isRouterReady) !== undefined
        ? adapter.isRouterReady
        : (contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.isRouterReady) !== undefined
            ? contextAdapter.isRouterReady
            : true;
    const urlQuery = (adapter === null || adapter === void 0 ? void 0 : adapter.urlQuery) || (contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.urlQuery);
    const onStateChange = (adapter === null || adapter === void 0 ? void 0 : adapter.onChange) || (contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.onChange);
    const mode = (adapter === null || adapter === void 0 ? void 0 : adapter.mode) || (contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.mode) || 'reset';
    const customSerializeQueryParam = (adapter === null || adapter === void 0 ? void 0 : adapter.customSerializeQueryParam) || (contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.customSerializeQueryParam);
    if (!onStateChange || !urlQuery) {
        throw new Error('useNextQueryParams is used outside a NextQueryParamsProvider, but no `onChange` function or `urlQuery` was passed to the hook.');
    }
    const uncontrolledKeys = Object.keys(urlQuery).filter((key) => !keys.includes(key));
    const queryParamStates = useMemo(() => {
        const queryParamStates = {};
        for (const [key, value] of Object.entries(params)) {
            queryParamStates[key] = {};
            if (value.value !== undefined && value.value !== null) {
                queryParamStates[key].value = value.value;
            }
            queryParamStates[key].serialize = value.serialize;
        }
        return queryParamStates;
    }, [params]);
    const hasUrlQueryChanged = useMemo(() => {
        return !areUrlQueriesEqual(urlQuery, prevUrlQuery);
    }, [urlQuery, prevUrlQuery]);
    const haveQueryParamStatesChanged = useMemo(() => {
        return !areUrlQueriesEqual(removeKeysFromObject(urlQuery, uncontrolledKeys), serializeQueryParamStates(queryParamStates, customSerializeQueryParam));
    }, [queryParamStates, urlQuery]);
    const onInit = () => {
        for (const key of keys) {
            const value = urlQuery[key];
            value && params[key].onChange(value);
        }
        prevHasUrlQueryChanged.current = true;
        setPrevUrlQuery(urlQuery);
        setInit(false);
    };
    const onChange = () => {
        var _a, _b;
        if (hasUrlQueryChanged) {
            const controlledChangedKeys = getChangedUrlQueryKeys(urlQuery, prevUrlQuery).filter((key) => keys.includes(key));
            for (const key of controlledChangedKeys) {
                const value = urlQuery[key];
                if (value === undefined) {
                    if (mode === 'reset') {
                        params[key].onReset();
                    }
                }
                else {
                    const serializedStateValue = params[key] && params[key].serialize
                        ? (_b = (_a = params[key]).serialize) === null || _b === void 0 ? void 0 : _b.call(_a, params[key].value)
                        : customSerializeQueryParam
                            ? customSerializeQueryParam(params[key].value)
                            : baseSerializeQueryParam(params[key].value);
                    if (serializedStateValue !== value) {
                        params[key].onChange(value);
                        prevHasUrlQueryChanged.current = true;
                    }
                }
            }
            setPrevUrlQuery(urlQuery);
        }
        else if (haveQueryParamStatesChanged) {
            onStateChange({
                ...removeKeysFromObject(urlQuery, keys),
                ...serializeQueryParamStates(queryParamStates, customSerializeQueryParam)
            }, prevHasUrlQueryChanged.current);
            prevHasUrlQueryChanged.current = false;
        }
        else {
            prevHasUrlQueryChanged.current = false;
        }
    };
    useEffect(() => {
        if (isRouterReady) {
            if (init) {
                onInit();
            }
            else {
                onChange();
            }
        }
    }, [init, isRouterReady, hasUrlQueryChanged, haveQueryParamStatesChanged]);
}

function createQueryParamFunctionFactory(implementation) {
    return function (props) {
        return implementation(props);
    };
}

const createStrQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize || ((v) => v),
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        if (Array.isArray(v)) {
            props.onChange(v[0]);
        }
        else {
            props.onChange(v);
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : '';
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

const createNumQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize,
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        if (Array.isArray(v)) {
            if (!isNaN(Number(v[0]))) {
                props.onChange(Number(v[0]));
            }
        }
        else {
            if (!isNaN(Number(v))) {
                props.onChange(Number(v));
            }
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : 0;
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

const createBoolQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize ||
        ((v) => {
            if (v === undefined || v === null) {
                return v;
            }
            return v.toString();
        }),
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        if (Array.isArray(v)) {
            if (v[0] === 'true') {
                props.onChange(true);
            }
            else if (v[0] === 'false') {
                props.onChange(false);
            }
        }
        else {
            if (v === 'true') {
                props.onChange(true);
            }
            else if (v === 'false') {
                props.onChange(false);
            }
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : false;
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

const createJSONRecordQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize ||
        ((v) => {
            if (v === undefined || v === null) {
                return v;
            }
            return JSON.stringify(v);
        }),
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        let parsed;
        try {
            if (Array.isArray(v)) {
                parsed = JSON.parse(v[0]);
            }
            else {
                parsed = JSON.parse(v);
            }
        }
        catch (e) {
            return;
        }
        if (isJSONRecord(parsed)) {
            props.onChange(parsed);
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : {};
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));
function isJSONRecord(value) {
    return typeof value === 'object' && value !== null && !Array.isArray(value);
}

const createDateQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize ||
        ((v) => {
            if (v === undefined || v === null) {
                return v;
            }
            if (props.withTime) {
                return v.toISOString().split('.')[0];
            }
            return v.toISOString().split('T')[0];
        }),
    onChange: (v) => {
        if (Array.isArray(v)) {
            const parsedDate = new Date(v[0]);
            if (!isNaN(parsedDate.getTime())) {
                props.onChange(parsedDate);
            }
        }
        else {
            const parsedDate = new Date(v);
            if (!isNaN(parsedDate.getTime())) {
                props.onChange(parsedDate);
            }
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue
            ? new Date(props.defaultValue)
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : new Date();
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

const createStrArrQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize,
    onChange: (v) => {
        if (Array.isArray(v)) {
            props.onChange(v);
        }
        else {
            props.onChange([v]);
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : [];
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

const createNumArrQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize,
    onChange: (v) => {
        if (Array.isArray(v)) {
            props.onChange(v.filter((v) => !isNaN(Number(v))).map((v) => Number(v)));
        }
        else {
            if (!isNaN(Number(v))) {
                props.onChange([Number(v)]);
            }
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : [];
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

function createNextRouterAdapter(nextRouter, options) {
    const { mode, replace, shallow, onChange } = options || {};
    const defaultOnChange = (query, isTriggeredByUrl) => {
        const updateMethod = replace || isTriggeredByUrl ? 'replace' : 'push';
        nextRouter[updateMethod]({
            pathname: nextRouter.pathname,
            query
        }, undefined, { shallow: !!shallow });
    };
    return {
        isRouterReady: nextRouter.isReady,
        urlQuery: nextRouter.query,
        mode,
        onChange: onChange || defaultOnChange
    };
}

export { NextQueryParamsProvider, createBoolQueryParam, createDateQueryParam, createJSONRecordQueryParam, createNextRouterAdapter, createNumArrQueryParam, createNumQueryParam, createQueryParamFunctionFactory, createStrArrQueryParam, createStrQueryParam, useNextQueryParams };
//# sourceMappingURL=index.esm.mjs.map
